#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <GL/glut.h>
#include <X11/Xlib.h>
#include <GL/glx.h>

#undef USE_SGI_SWAP_CONTROL

int vsync_control(int vsync);
void disp(void);
void reshape(int x, int y);
void keyb(unsigned char key, int x, int y);

int main(int argc, char **argv)
{
	int vsync = 0;
	float ldir[] = {-1, 1, 1, 0};

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowSize(640, 480);
	glutCreateWindow("test");

	glutDisplayFunc(disp);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutIdleFunc(glutPostRedisplay);

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_POSITION, ldir);

	if(argv[1] && strcmp(argv[1], "-vsync") == 0) {
		vsync = 1;
	}
	if(vsync_control(vsync) == -1) {
		fprintf(stderr, "failed to control vsync\n");
	} else {
		printf("vsync %s\n", vsync ? "enabled" : "disabled");
	}

	glutMainLoop();
	return 0;
}

int vsync_control(int vsync)
{
	unsigned int interval;
#ifdef USE_SGI_SWAP_CONTROL
	void (*glXSwapIntervalSGI)(int);
#else
	Display *dpy;
	GLXDrawable win;
	void (*glXSwapIntervalEXT)(Display*, Drawable, int);
#endif

	if(vsync) vsync = 1;

#ifdef USE_SGI_SWAP_CONTROL
	if(!(glXSwapIntervalSGI = glXGetProcAddress((unsigned char*)"glXSwapIntervalSGI"))) {
		fprintf(stderr, "missing extension: GLX_SGI_swap_control\n");
		return -1;
	}

	glXSwapIntervalSGI(vsync);
#else
	dpy = glXGetCurrentDisplay();
	assert(dpy);
	win = glXGetCurrentDrawable();
	assert(win);

	if(!(glXSwapIntervalEXT = glXGetProcAddress((unsigned char*)"glXSwapIntervalEXT"))) {
		fprintf(stderr, "missing extension: GLX_EXT_swap_control\n");
		return -1;
	}

	/* enable vsync */
	glXSwapIntervalEXT(dpy, win, vsync);

	/* sanity check */
	glXQueryDrawable(dpy, win, GLX_SWAP_INTERVAL_EXT, &interval);
	if(interval != vsync) {
		fprintf(stderr, "sanity check failed\n");
		return -1;
	}
#endif
	return 0;
}

void disp(void)
{
	static int frames;
	static unsigned long start_time;
	unsigned long msec = glutGet(GLUT_ELAPSED_TIME);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0, 0, -6);
	glRotatef(30.0, 1, 0, 0);
	glRotatef(msec / 10.0, 0, 1, 0);

	glFrontFace(GL_CW);
	glutSolidTeapot(1.0);
	glFrontFace(GL_CCW);

	glutSwapBuffers();
	assert(glGetError() == 0);
	frames++;

	msec = glutGet(GLUT_ELAPSED_TIME);
	if(msec - start_time >= 2000) {
		char buf[64];
		float fps = 1000.0 * frames / (msec - start_time);

		sprintf(buf, "framerate: %.2f fps", fps);
		glutSetWindowTitle(buf);
		printf("\r%s\t", buf);
		fflush(stdout);

		start_time = msec;
		frames = 0;
	}
}

void reshape(int x, int y)
{
	glViewport(0, 0, x, y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, (float)x / (float)y, 1.0, 1000.0);
}

void keyb(unsigned char key, int x, int y)
{
	if(key == 27) {
		putchar('\n');
		exit(0);
	}
}
